# pylint: disable=redefined-outer-name,missing-module-docstring,missing-function-docstring
# pylint: disable=import-outside-toplevel
import datetime as dt

import pytest
from dask.base import tokenize
from odc.geo.geobox import GeoBox
from odc.loader.types import (
    AuxDataSource,
    BandKey,
    RasterBandMetadata,
    RasterGroupMetadata,
    RasterLoadParams,
    RasterSource,
    norm_key,
)

from odc.stac import ParsedItem, RasterCollectionMetadata
from odc.stac.model import PropertyLoadRequest
from odc.stac.testing.stac import b_, mk_parsed_item


def test_band_load_info() -> None:
    meta = RasterBandMetadata(data_type="uint16", nodata=13)
    band = RasterSource("https://example.com/some.tif", meta=meta)
    assert RasterLoadParams.same_as(meta).dtype == "uint16"
    assert RasterLoadParams.same_as(band).fill_value == 13

    band = RasterSource("file:///")
    assert RasterLoadParams.same_as(band).dtype == "float32"
    assert RasterLoadParams().dtype is None
    assert RasterLoadParams().nearest is True
    assert RasterLoadParams(resampling="average").nearest is False


@pytest.mark.parametrize("lon", [0, -179, 179, 10, 23.4])
def test_mid_longitude(lon: float) -> None:
    gbox = GeoBox.from_bbox((lon - 0.1, 0, lon + 0.1, 1), shape=(100, 100))
    xx = mk_parsed_item([b_("b1", gbox)])
    assert xx.geometry is not None
    assert xx.geometry.crs == "epsg:4326"
    assert xx.mid_longitude == pytest.approx(lon)

    assert mk_parsed_item([]).mid_longitude is None


def test_solar_day() -> None:
    def _mk(lon: float, datetime):
        gbox = GeoBox.from_bbox((lon - 0.1, 0, lon + 0.1, 1), shape=(100, 100))
        return mk_parsed_item([b_("b1", gbox)], datetime=datetime)

    for lon in [0, 1, 2, 3, 14, -1, -14, -3]:
        xx = _mk(lon, "2020-01-02T12:13:14Z")
        assert xx.mid_longitude == pytest.approx(lon)
        assert xx.nominal_datetime == xx.solar_date

    xx = _mk(15.1, "2020-01-02T12:13:14Z")
    assert xx.nominal_datetime != xx.solar_date
    assert xx.nominal_datetime + dt.timedelta(seconds=3600) == xx.solar_date
    assert xx.nominal_datetime + dt.timedelta(seconds=3600) == xx.solar_date_at(20)

    xx = _mk(-15.1, "2020-01-02T12:13:14Z")
    assert xx.nominal_datetime != xx.solar_date
    assert xx.nominal_datetime - dt.timedelta(seconds=3600) == xx.solar_date
    assert xx.nominal_datetime - dt.timedelta(seconds=3600) == xx.solar_date_at(-20)

    xx = mk_parsed_item([b_("b1")], datetime="2000-01-02")
    assert xx.geometry is None
    assert xx.nominal_datetime == xx.solar_date

    xx = _mk(10, None)
    with pytest.raises(ValueError):
        _ = xx.solar_date


@pytest.fixture()
def collection_ab() -> RasterCollectionMetadata:
    return RasterCollectionMetadata(
        "ab",
        RasterGroupMetadata(
            {
                ("a", 1): RasterBandMetadata("uint8"),
                ("b", 1): RasterBandMetadata("uint16"),
            },
            {"A": [("a", 1)], "AA": [("a", 1)], "B": [("b", 1)]},
        ),
        has_proj=True,
        band2grid={},
    )


@pytest.fixture()
def parsed_item_ab(collection_ab: RasterCollectionMetadata) -> ParsedItem:
    def _src(k: BandKey) -> RasterSource | AuxDataSource:
        meta = collection_ab[k]
        if isinstance(meta, RasterBandMetadata):
            return RasterSource(f"file:///{k[0]}-{k[1]}.tif", meta=meta)
        return AuxDataSource(f"file:///{k[0]}-{k[1]}.aux", meta=meta)

    return ParsedItem(
        "item-ab",
        collection_ab,
        {k: _src(k) for k in collection_ab},
    )


def test_collection(collection_ab: RasterCollectionMetadata) -> None:
    xx = collection_ab

    assert xx.canonical_name("b") == "b"
    assert xx.canonical_name("B") == "b"
    assert xx.canonical_name("AA") == "a"
    assert xx.canonical_name("a") == "a"

    assert xx.band_key("B") == ("b", 1)
    assert xx.band_key("AA") == ("a", 1)
    assert xx["AA"].data_type == "uint8"
    assert xx["b"].data_type == "uint16"
    assert "b" in xx
    assert "b.1" in xx
    assert ("b", 1) in xx
    assert {} not in xx
    assert ("some-random", 1) not in xx
    assert "no-such-band" not in xx

    assert xx.resolve_bands("AA")["AA"] == xx["a"]
    assert list(xx.resolve_bands(["a", "B"])) == ["a", "B"]
    assert xx.resolve_bands(["a", "B"])["B"] is xx["b"]
    assert xx.resolve_bands(["a", "B"])["a"] is xx["a"]
    assert set(xx) == set([("a", 1), ("b", 1)])
    assert len(xx) == 2

    for k in "a AA A b B".split(" "):
        assert xx.band_key(k) in xx.bands
        assert xx.canonical_name(k) in ["a", "b"]
        assert k in xx
        assert isinstance(xx[k], RasterBandMetadata)
        assert xx[k] is xx[xx.band_key(k)]

    with pytest.raises(ValueError):
        _ = xx.resolve_bands(["xxxxxxxx", "a"])

    with pytest.raises(KeyError):
        _ = xx["no-such-band"]


def test_collection_allbands() -> None:
    xx = mk_parsed_item([b_("a.1"), b_("a.2"), b_("a.3")])
    md = xx.collection
    assert md.all_bands == ["a.1", "a.2", "a.3"]

    md.aliases["AA"] = [("a", 2)]
    md.aliases["AAA"] = [("a", 3)]
    assert md["AA"] == md["a.2"]
    assert md["AAA"] == md["a.3"]

    # expect aliases to be used for all_band when multi-band
    # assets have unique aliases
    assert md.all_bands == ["a.1", "AA", "AAA"]
    assert md.canonical_name("a.2") == "AA"
    assert md.canonical_name("AA") == "AA"
    assert md.canonical_name("a.3") == "AAA"
    assert md.canonical_name("AAA") == "AAA"


def test_parsed_item(parsed_item_ab: ParsedItem) -> None:
    xx = parsed_item_ab
    assert xx["AA"] is not None
    assert xx["b"] is not None
    assert xx["AA"].meta is not None
    assert xx["AA"].meta.data_type == "uint8"
    assert xx["b"].meta is not None
    assert xx["b"].meta.data_type == "uint16"

    assert xx.resolve_bands("AA")["AA"] == xx["a"]
    assert list(xx.resolve_bands(["a", "B"])) == ["a", "B"]
    assert xx.resolve_bands(["a", "B"])["B"] is xx["b"]
    assert xx.resolve_bands(["a", "B"])["a"] is xx["a"]
    assert set(xx) == set([("a", 1), ("b", 1)])
    assert len(xx) == 2
    assert len(set([xx, xx, xx])) == 1
    assert ("a", 1) in xx
    assert ("a", 2) not in xx
    assert ("a", 2, 3) not in xx

    for k in "a AA A b B".split(" "):
        assert k in xx
        assert [k] not in xx
        assert f"___{k}___" not in xx
        assert isinstance(xx[k], RasterSource)
        assert xx[k] is xx.resolve_bands(k)[k]

    assert isinstance(xx["b"], RasterSource)
    assert isinstance(xx["b"].strip(), RasterSource)
    assert xx["b"].strip().geobox is None
    assert xx["b"].strip().meta is xx["b"].meta
    assert xx["b"].strip().uri == xx["b"].uri
    assert xx["b"].strip().band == xx["b"].band
    assert xx["b"].strip().subdataset == xx["b"].subdataset
    assert xx["b"].strip().driver_data == xx["b"].driver_data

    xx_strip = xx.strip()
    assert isinstance(xx_strip["b"], RasterSource)
    assert xx_strip["b"].geobox is None
    assert xx_strip["b"].meta is xx["b"].meta
    assert xx_strip["b"].uri == xx["b"].uri
    assert xx_strip["b"].band == xx["b"].band
    assert xx_strip["b"].subdataset == xx["b"].subdataset
    assert xx_strip["b"].driver_data == xx["b"].driver_data


def test_tokenize(parsed_item_ab: ParsedItem) -> None:
    assert tokenize(parsed_item_ab.collection) == tokenize(parsed_item_ab.collection)
    assert tokenize(parsed_item_ab) == tokenize(parsed_item_ab)
    assert tokenize(parsed_item_ab["a"]) == tokenize(parsed_item_ab["a"])
    assert tokenize(parsed_item_ab["a"].meta) == tokenize(parsed_item_ab["a"].meta)

    assert tokenize(RasterLoadParams()) == tokenize(RasterLoadParams())
    assert tokenize(RasterLoadParams("uint8")) == tokenize(RasterLoadParams("uint8"))
    assert tokenize(RasterLoadParams("uint8")) != tokenize(RasterLoadParams("uint32"))


@pytest.mark.parametrize(
    "name, expected",
    [
        ("a", ("a", 1)),
        ("a.1", ("a", 1)),
        ("a.2", ("a", 2)),
        (("b", 1), ("b", 1)),
        ("foo.tiff", ("foo.tiff", 1)),
    ],
)
def test_normkey(name, expected) -> None:
    assert norm_key(name) == expected


def test_version() -> None:
    from odc.stac import __version__  # pylint: disable=no-name-in-module

    assert __version__ is not None
    assert len(__version__.split(".")) == 3


def test_property_load_request_basic() -> None:
    """Test basic PropertyLoadRequest functionality."""
    # Test with just key
    req = PropertyLoadRequest(key="eo:cloud_cover")
    assert req.key == "eo:cloud_cover"
    assert req.dtype == "float32"  # default
    assert req.name is None  # default
    assert req.nodata is None
    assert req.units == "1"
    assert req.output_name == "eo_cloud_cover"

    # Test with all fields
    req = PropertyLoadRequest(
        key="eo:cloud_cover", dtype="int16", name="cloud_cover", nodata=-999
    )
    assert req.key == "eo:cloud_cover"
    assert req.dtype == "int16"
    assert req.name == "cloud_cover"
    assert req.nodata == -999
    assert req.units == "1"
    assert req.output_name == "cloud_cover"


def test_property_load_request_from_user_input() -> None:
    """Test from_user_input method with various inputs."""
    # Test with string inputs
    requests = PropertyLoadRequest.from_user_input(["eo:cloud_cover", "eo:platform"])
    assert len(requests) == 2
    assert requests[0].key == "eo:cloud_cover"
    assert requests[1].key == "eo:platform"
    assert all(req.dtype == "float32" for req in requests)
    assert all(req.name is None for req in requests)

    # Test with dict inputs
    requests = PropertyLoadRequest.from_user_input(
        [
            {"key": "eo:cloud_cover", "dtype": "int16", "name": "cloud_cover"},
            {"key": "eo:platform", "name": "satellite"},
        ]
    )
    assert len(requests) == 2
    assert requests[0].key == "eo:cloud_cover"
    assert requests[0].dtype == "int16"
    assert requests[0].name == "cloud_cover"
    assert requests[1].key == "eo:platform"
    assert requests[1].dtype == "float32"  # default
    assert requests[1].name == "satellite"

    # Test with mixed inputs
    requests = PropertyLoadRequest.from_user_input(
        ["eo:cloud_cover", {"key": "eo:platform", "name": "satellite"}]
    )
    assert len(requests) == 2
    assert requests[0].key == "eo:cloud_cover"
    assert requests[0].dtype == "float32"
    assert requests[0].name is None
    assert requests[1].key == "eo:platform"
    assert requests[1].name == "satellite"


def test_property_load_request_errors() -> None:
    """Test error cases for PropertyLoadRequest."""
    # Test missing key in dict
    with pytest.raises(ValueError, match="Dictionary input must contain 'key' field"):
        PropertyLoadRequest.from_user_input([{"dtype": "int16"}])

    # Test invalid input type
    with pytest.raises(ValueError, match="Input must be string or dict"):
        PropertyLoadRequest.from_user_input([123])  # type: ignore

    # Test empty sequence
    requests = PropertyLoadRequest.from_user_input([])
    assert len(requests) == 0



import hashlib
import os.path


def compute(folder: str) -> str:
    hash = hashlib.sha256()
    paths = [
        os.path.join(folder, file_name)
        for file_name in os.listdir(folder)
        if os.path.splitext(file_name)[1] == ".py"
    ]
    paths = sorted(paths, key=str.casefold)
    for path in paths:
        with open(path, "rb") as file:
            bytes = file.read()
            hash.update(bytes)
    return hash.hexdigest(), paths


if __name__ == "__main__":
    folder = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "notebooks"))
    hsh, _ = compute(folder)
    print(hsh)



"""Helper methods for benchmark reporting."""

import glob
import pickle
from typing import Any, Dict, Iterable, Iterator, Union

import pandas as pd

# pylint: disable=unsupported-assignment-operation


def load_results(
    sources: Union[str, Iterable[str]],
) -> pd.DataFrame:
    """
    Load benchmark run results.

    :param sources: A glob pattern or a stream of pickle file paths
    :return: Pandas dataframe
    """

    def _stream(paths: Iterable[str]) -> Iterator[Dict[str, Any]]:
        for idx, fname in enumerate(paths):
            with open(fname, "rb") as src:
                dd = pickle.load(src)
            ctx = dd["context"]
            samples = dd["samples"]
            rr = ctx.to_pandas_dict()

            for sample in samples:
                t0, t1, t2 = sample
                yield {"experiment": idx, **rr, "t0": t0, "t1": t1, "t2": t2}

    if isinstance(sources, str):
        # glob
        pkl_paths: Iterable[str] = sorted(glob.glob(sources))
    else:
        pkl_paths = sources

    xx = pd.DataFrame(list(_stream(pkl_paths)))
    xx = xx.set_index("experiment")
    xx["submit"] = xx.t1 - xx.t0
    xx["elapsed"] = xx.t2 - xx.t0
    return xx



"""
Making STAC items for testing.
"""

from datetime import datetime, timezone
from typing import Any, Generator

import pystac.asset
import pystac.item
import xarray as xr
from odc.geo.geobox import GeoBox
from odc.loader.types import (
    AuxBandMetadata,
    AuxDataSource,
    RasterBandMetadata,
    RasterGroupMetadata,
    RasterSource,
    norm_key,
)
from pystac.extensions.projection import ProjectionExtension
from pystac.extensions.raster import RasterBand, RasterExtension
from toolz import dicttoolz

from .._mdtools import _group_geoboxes
from ..model import ParsedItem, PropertyLoadRequest, RasterCollectionMetadata

# pylint: disable=redefined-builtin,too-many-arguments

STAC_DATE_FMT = "%Y-%m-%dT%H:%M:%S.%fZ"
STAC_DATE_FMT_SHORT = "%Y-%m-%dT%H:%M:%SZ"


def _norm_dates(*args):
    valid = [a for a in args if a is not None]
    valid = [
        datetime.fromisoformat(dt).replace(tzinfo=timezone.utc)
        for dt in xr.DataArray(list(valid))
        .astype("datetime64[ns]")
        .dt.strftime("%Y-%m-%dT%H:%M:%S.%f")
        .values
    ]
    valid = iter(valid)
    return [next(valid) if a else None for a in args]


def b_(
    name,
    geobox=None,
    dtype="int16",
    nodata=None,
    unit="1",
    dims=(),
    uri=None,
    bidx=1,
    prefix="http://example.com/items/",
):
    band_key = norm_key(name)
    name, _ = band_key
    if uri is None:
        uri = f"{prefix}{name}.tif"
    meta = RasterBandMetadata(dtype, nodata, unit, dims=dims)
    return (band_key, RasterSource(uri, bidx, geobox=geobox, meta=meta))


def mk_parsed_item(
    bands,
    datetime=None,
    *,
    start_datetime=None,
    end_datetime=None,
    id="some-item",
    collection="some-collection",
    href=None,
    geometry=None,
    props: dict[str, Any] | None = None,
) -> ParsedItem:
    """
    Construct parsed stac item for testing.
    """
    # pylint: disable=redefined-outer-name, too-many-locals
    if isinstance(bands, (list, tuple)):
        bands = {norm_key(k): v for k, v in bands}

    gboxes = dicttoolz.valmap(lambda b: b.geobox, bands)
    gboxes = dicttoolz.valfilter(lambda x: x is not None, gboxes)
    gboxes = dicttoolz.keymap(lambda bk: bk[0], gboxes)

    if len(gboxes) == 0:
        band2grid = {b: "default" for b, _ in bands}
        geobox = None
    else:
        grids, band2grid = _group_geoboxes(gboxes)
        geobox = grids["default"]

    if geometry is None and geobox is not None:
        geometry = geobox.geographic_extent

    aliases = {}
    if props is None:
        props = {}

    # Handle auxiliary bands from props
    prop_user_input = [v[1] if isinstance(v, tuple) else k for k, v in props.items()]
    prop_requests = PropertyLoadRequest.from_user_input(prop_user_input)
    for idx, prop_req in enumerate(prop_requests):
        bk = ("_stac_metadata", idx + 1)
        # Look up actual value from props dict using prop_req.key
        actual_value = props[prop_req.key]
        if isinstance(actual_value, tuple):
            actual_value, _ = actual_value

        aux_meta = AuxBandMetadata(
            prop_req.dtype,
            nodata=prop_req.nodata,
            units=prop_req.units,
            driver_data=prop_req,
        )
        aux_source = AuxDataSource(
            uri=f"virtual://{bk[0]}/{bk[1]}",
            subdataset=None,
            meta=aux_meta,
            driver_data=actual_value,
        )
        bands[bk] = aux_source
        aliases[prop_req.output_name] = [bk]

    collection = RasterCollectionMetadata(
        collection,
        RasterGroupMetadata(
            dicttoolz.valmap(lambda b: b.meta, bands),
            aliases=aliases,
        ),
        has_proj=(geobox is not None),
        band2grid=band2grid,
    )
    datetime, start_datetime, end_datetime = _norm_dates(
        datetime, start_datetime, end_datetime
    )

    return ParsedItem(
        id,
        collection,
        bands,
        geometry=geometry,
        datetime=datetime,
        datetime_range=(start_datetime, end_datetime),
        href=href,
    )


def _add_proj(gbox: GeoBox, xx) -> None:
    proj = ProjectionExtension.ext(xx, add_if_missing=True)
    proj.shape = list(gbox.shape.yx)
    proj.transform = gbox.transform[:6]
    crs = gbox.crs
    if crs is not None:
        epsg = crs.epsg
        if epsg is not None:
            proj.epsg = epsg
        else:
            proj.wkt2 = crs.wkt


def _extract_props(item: ParsedItem) -> Generator[tuple[str, Any], None, None]:
    for k in item.bands:
        if k[0] != "_stac_metadata":
            continue
        b = item[k]
        if b.meta is None or b.meta.driver_data is None:
            continue
        yield b.meta.driver_data.key, b.driver_data


def to_stac_item(item: ParsedItem) -> pystac.item.Item:
    gg = item.geometry

    props = {}
    for n, dt in zip(["start_datetime", "end_datetime"], item.datetime_range):
        if dt is not None:
            props[n] = dt.strftime(STAC_DATE_FMT)

    props.update(_extract_props(item))

    xx = pystac.item.Item(
        item.id,
        geometry=gg.json if gg is not None else None,
        bbox=list(gg.boundingbox.bbox) if gg is not None else None,
        datetime=item.datetime,
        properties=props,
        collection=item.collection.name,
    )

    gboxes = item.geoboxes()
    if len(gboxes) > 0:
        gbox = gboxes[0]

        ProjectionExtension.add_to(xx)
        _add_proj(gbox, xx)

    def _to_raster_band(src: RasterSource) -> RasterBand:
        meta = src.meta
        assert meta is not None
        return RasterBand.create(
            data_type=meta.data_type,  # type: ignore[arg-type]
            nodata=meta.nodata,
            unit=meta.units,
        )

    for asset_name, bands in item.assets().items():
        RasterExtension.add_to(xx)
        b = bands[0]  # all bands should share same uri
        xx.add_asset(
            asset_name,
            pystac.asset.Asset(b.uri, media_type="image/tiff", roles=["data"]),
        )
        RasterExtension.ext(xx.assets[asset_name]).apply(
            list(map(_to_raster_band, bands))
        )

    for asset_name, asset in xx.assets.items():
        bb = item.bands[(asset_name, 1)]
        assert isinstance(bb, RasterSource)
        if bb.geobox is not None:
            assert isinstance(bb.geobox, GeoBox)
            _add_proj(bb.geobox, asset)

    if item.href is not None:
        xx.set_self_href(item.href)

    return xx



# pylint: disable=wrong-import-order,wrong-import-position,
# pylint: disable=redefined-outer-name,missing-function-docstring,missing-module-docstring
import pytest

distributed = pytest.importorskip("distributed")

from unittest.mock import MagicMock

import xarray
import numpy as np
from distributed import Client
from odc.geo.xr import ODCExtension

from odc.stac.bench import (
    BenchLoadParams,
    collect_context_info,
    load_from_json,
    run_bench,
)

CFG = {
    "*": {
        "warnings": "ignore",
        # for every asset in every product default to uint16 with nodata=0
        "assets": {"*": {"data_type": "uint16", "nodata": 0}},
    }
}


@pytest.fixture
def fake_dask_client(monkeypatch):
    cc = MagicMock()
    cc.scheduler_info.return_value = {
        "type": "Scheduler",
        "id": "Scheduler-80d943db-16f6-4476-a51a-64d57a287e9b",
        "address": "inproc://10.10.10.10/1281505/1",
        "services": {"dashboard": 8787},
        "started": 1638320006.6135786,
        "workers": {
            "inproc://10.10.10.10/1281505/4": {
                "type": "Worker",
                "id": 0,
                "host": "10.1.1.140",
                "resources": {},
                "local_directory": "/tmp/dask-worker-space/worker-uhq1b9bh",
                "name": 0,
                "nthreads": 2,
                "memory_limit": 524288000,
                "last_seen": 1638320007.2504623,
                "services": {"dashboard": 38439},
                "metrics": {
                    "executing": 0,
                    "in_memory": 0,
                    "ready": 0,
                    "in_flight": 0,
                    "bandwidth": {"total": 100000000, "workers": {}, "types": {}},
                    "spilled_nbytes": 0,
                    "cpu": 0.0,
                    "memory": 145129472,
                    "time": 1638320007.2390554,
                    "read_bytes": 0.0,
                    "write_bytes": 0.0,
                    "read_bytes_disk": 0.0,
                    "write_bytes_disk": 0.0,
                    "num_fds": 82,
                },
                "nanny": None,
            }
        },
    }
    cc.cancel.return_value = None
    cc.restart.return_value = cc
    cc.persist = lambda x: x
    cc.compute = lambda x: x

    monkeypatch.setattr(distributed, "wait", MagicMock())
    yield cc


@pytest.fixture(scope="module")
def dask_client():
    client = Client(
        n_workers=1,
        threads_per_worker=2,
        memory_limit="500MiB",
        local_directory="/tmp/",
        memory_target_fraction=False,
        memory_spill_fraction=False,
        memory_pause_fraction=False,
        dashboard_address=None,
        processes=False,
    )
    yield client
    client.shutdown()
    del client


@pytest.mark.skipif(
    not pytest.importorskip("stackstac"), reason="stackstac not installed"
)
def test_load_from_json_stackstac(fake_dask_client, bench_site1, bench_site2) -> None:
    dask_client = fake_dask_client
    params = BenchLoadParams(
        scenario="test1",
        method="stackstac",
        bands=("B04", "B02", "B03"),
        chunks=(2048, 2048),
        resampling="nearest",
        extra={
            "odc-stac": {"groupby": "solar_day", "stac_cfg": CFG},
            "stackstac": {
                "dtype": "uint16",
                "fill_value": np.uint16(0),
                "rescale": False,
            },
        },
    )
    xx = load_from_json(bench_site1, params)
    assert "band" in xx.dims
    assert xx.shape == (1, 3, 90978, 10980)
    assert xx.dtype == "uint16"
    assert xx.spec.epsg == 32735

    yy = load_from_json(
        bench_site1, params.with_method("odc-stac"), geobox=xx.odc.geobox
    )

    rrx = collect_context_info(dask_client, xx)
    rry = collect_context_info(dask_client, yy)
    assert rrx.shape == rry.shape
    assert rrx == rry

    xx = load_from_json(bench_site2, params)
    assert "band" in xx.dims
    assert xx.dtype == "uint16"
    assert xx.spec.epsg == 32735

    params.crs = "epsg:32736"
    xx = load_from_json(bench_site2, params)
    assert "band" in xx.dims
    assert xx.dtype == "uint16"
    assert xx.spec.epsg == 32736

    with pytest.raises(ValueError):
        load_from_json(bench_site1, params.with_method("wroNg"))


def test_bench_context(fake_dask_client, bench_site1, bench_site2) -> None:
    params = BenchLoadParams(
        scenario="test1",
        method="odc-stac",
        bands=("red", "green", "blue"),
        chunks=(2048, 2048),
        extra={"odc-stac": {"groupby": "solar_day", "stac_cfg": CFG}},
    )
    xx = load_from_json(bench_site1, params)
    nt, ny, nx = xx.red.shape
    nb = len(xx.data_vars)

    # Check normal case Dataset, with time coords
    rr = collect_context_info(
        fake_dask_client, xx, method=params.method, scenario="site1"
    )
    assert isinstance(xx.odc, ODCExtension)
    assert rr.shape == (nt, nb, ny, nx)
    assert rr.chunks == (1, 1, 2048, 2048)
    assert rr.crs == f"epsg:{xx.odc.geobox.crs.epsg}"
    assert rr.crs == xx.odc.geobox.crs
    assert rr.nthreads == 2
    assert rr.total_ram == 500 * (1 << 20)

    header_txt = rr.render_txt()
    assert "T.slice   : 2020-06-06" in header_txt
    assert f"Data      : 1.3.{ny}.{nx}.uint16,  5.58 GiB" in header_txt

    run_txt = rr.render_timing_info((0, 0.1, 30))
    assert isinstance(run_txt, str)

    pd_dict = rr.to_pandas_dict()
    assert pd_dict["resolution"] == rr.resolution
    assert pd_dict["data"] == f"1.3.{ny}.{nx}.uint16"
    assert pd_dict["chunks_x"] == 2048
    assert pd_dict["chunks_y"] == 2048

    # Check DataArray case
    rr = collect_context_info(
        fake_dask_client, xx.red, method="odc-stac", scenario="site1"
    )
    assert rr.shape == (nt, 1, ny, nx)
    assert rr.crs == xx.odc.geobox.crs

    # Check Dataset with 0 dimension time axis and extras field
    rr = collect_context_info(
        fake_dask_client,
        xx.isel(time=0),
        method=params.method,
        scenario=params.scenario,
        extras={"custom": 2},
    )
    assert rr.extras == {"custom": 2}
    assert rr.shape == (1, nb, ny, nx)

    header_txt = rr.render_txt()
    assert "GEO       : epsg:32735" in header_txt
    assert "T.slice   : 2020-06-06" in header_txt

    # Check no time info at all
    rr = collect_context_info(
        fake_dask_client,
        xx.isel(time=0, drop=True),
        method=params.method,
        scenario=params.scenario,
    )
    assert rr.shape == (nt, nb, ny, nx)
    assert rr.dtype == xx.red.dtype
    assert rr.temporal_id == "-"

    # Check wrong type
    with pytest.raises(ValueError):
        collect_context_info(fake_dask_client, "wrong input type")  # type: ignore

    # Check multi-time axis
    xx = load_from_json(bench_site2, params)
    nt, ny, nx = xx.red.shape
    nb = len(xx.data_vars)

    assert nt > 1

    rr = collect_context_info(
        fake_dask_client,
        xx,
        method=params.method,
        scenario=params.scenario,
    )
    assert rr.shape == (nt, nb, ny, nx)
    assert rr.temporal_id == "2020-06-01__2020-07-31"

    # Check missing GEO info
    no_geo = _strip_geo(xx.red)
    assert no_geo.odc.geobox is None or no_geo.odc.geobox.crs is None
    with pytest.raises(ValueError):
        # no geobox
        collect_context_info(fake_dask_client, no_geo)


def _strip_geo(xx: xarray.DataArray) -> xarray.DataArray:
    no_geo = xx.drop_vars("spatial_ref")
    no_geo.attrs.pop("crs", None)
    no_geo.attrs.pop("grid_mapping", None)
    no_geo.encoding.pop("grid_mapping", None)
    no_geo.x.attrs.pop("crs", None)
    no_geo.y.attrs.pop("crs", None)
    # get rid of cached geobox
    no_geo = xarray.DataArray(
        no_geo.data,
        coords=no_geo.coords,
        dims=no_geo.dims,
        attrs=no_geo.attrs,
    )
    assert no_geo.odc.geobox is None or no_geo.odc.geobox.crs is None
    return no_geo


def test_run_bench(fake_dask_client, bench_site1, capsys) -> None:
    dask_client = fake_dask_client
    params = BenchLoadParams(
        scenario="test1",
        method="odc-stac",
        bands=("red", "green", "blue"),
        chunks=(2048, 2048),
        extra={"odc-stac": {"groupby": "solar_day", "stac_cfg": CFG}},
    )
    xx = load_from_json(bench_site1, params)

    rr, timing = run_bench(xx, dask_client, 10)

    assert rr.scenario == params.scenario
    assert rr.method == params.method
    assert len(timing) == 10
    _io = capsys.readouterr()
    assert len(_io.out) > 0


def test_bench_params_json() -> None:
    params = BenchLoadParams(
        scenario="test1",
        method="odc-stac",
        bands=("red", "green", "blue"),
        chunks=(100, 200),
        extra={"odc-stac": {"groupby": "solar_day", "stac_cfg": CFG}},
    )

    assert params == BenchLoadParams.from_json(params.to_json())
    assert params.to_json() == BenchLoadParams.from_json(params.to_json()).to_json()

    # function should round-trip too
    params.patch_url = load_from_json
    assert params == BenchLoadParams.from_json(params.to_json())



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.13.8
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Access Sentinel 2 Data from AWS
#
# [![Binder](https://mybinder.org/badge_logo.svg)](https://mybinder.org/v2/gh/opendatacube/odc-stac/develop?labpath=notebooks%2Fstac-load-e84-aws.ipynb)
#
# https://registry.opendata.aws/sentinel-2-l2a-cogs/

# %%
import dask.distributed
import folium
import folium.plugins
import geopandas as gpd
import shapely.geometry
from IPython.display import display
from pystac_client import Client

from odc.stac import configure_rio, stac_load


def convert_bounds(bbox, invert_y=False):
    """
    Helper method for changing bounding box representation to leaflet notation

    ``(lon1, lat1, lon2, lat2) -> ((lat1, lon1), (lat2, lon2))``
    """
    x1, y1, x2, y2 = bbox
    if invert_y:
        y1, y2 = y2, y1
    return ((y1, x1), (y2, x2))


# %% [markdown]
# ## Start Dask Client
#
# This step is optional, but it does improve load speed significantly. You
# don't have to use Dask, as you can load data directly into memory of the
# notebook.

# %%
client = dask.distributed.Client()
configure_rio(cloud_defaults=True, aws={"aws_unsigned": True}, client=client)
display(client)

# %% [markdown]
# ## Find STAC Items to Load

# %%
km2deg = 1.0 / 111
x, y = (113.887, -25.843)  # Center point of a query
r = 100 * km2deg
bbox = (x - r, y - r, x + r, y + r)

catalog = Client.open("https://earth-search.aws.element84.com/v1/")

query = catalog.search(
    collections=["sentinel-2-l2a"], datetime="2021-09-16", limit=100, bbox=bbox
)

items = list(query.items())
print(f"Found: {len(items):d} datasets")

# Convert STAC items into a GeoJSON FeatureCollection
stac_json = query.item_collection_as_dict()

# %% [markdown]
# ## Review Query Result
#
# We'll use GeoPandas DataFrame object to make plotting easier.

# %%
gdf = gpd.GeoDataFrame.from_features(stac_json, "epsg:4326")

# Compute granule id from components
gdf["granule"] = (
    gdf["mgrs:utm_zone"].apply(lambda x: f"{x:02d}")
    + gdf["mgrs:latitude_band"]
    + gdf["mgrs:grid_square"]
)

fig = gdf.plot(
    "granule",
    edgecolor="black",
    categorical=True,
    aspect="equal",
    alpha=0.5,
    figsize=(6, 12),
    legend=True,
    legend_kwds={"loc": "upper left", "frameon": False, "ncol": 1},
)
_ = fig.set_title("STAC Query Results")

# %% [markdown]
# ## Plot STAC Items on a Map

# %%
# https://github.com/python-visualization/folium/issues/1501
from branca.element import Figure

fig = Figure(width="400px", height="500px")
map1 = folium.Map()
fig.add_child(map1)

folium.GeoJson(
    shapely.geometry.box(*bbox),
    style_function=lambda x: dict(fill=False, weight=1, opacity=0.7, color="olive"),
    name="Query",
).add_to(map1)

gdf.explore(
    "granule",
    categorical=True,
    tooltip=[
        "granule",
        "datetime",
        "eo:cloud_cover",
    ],
    popup=True,
    style_kwds=dict(fillOpacity=0.1, width=2),
    name="STAC",
    m=map1,
)

map1.fit_bounds(bounds=convert_bounds(gdf.unary_union.bounds))
display(fig)

# %% [markdown]
# ## Construct Dask Dataset
#
# Note that even though there are 9 STAC Items on input, there is only one
# timeslice on output. This is because of `groupby="solar_day"`. With that
# setting `stac_load` will place all items that occured on the same day (as
# adjusted for the timezone) into one image plane.

# %%
# Since we will plot it on a map we need to use `EPSG:3857` projection
crs = "epsg:3857"
zoom = 2**5  # overview level 5

xx = stac_load(
    items,
    bands=("red", "green", "blue"),
    crs=crs,
    resolution=10 * zoom,
    chunks={},  # <-- use Dask
    groupby="solar_day",
)
display(xx)

# %% [markdown]
# Note that data is not loaded yet. But we can review memory requirement. We can also check data footprint.

# %%
xx.odc.geobox

# %% [markdown]
# ## Load data into local memory

# %%
# %%time
xx = xx.compute()

# %%
_ = (
    xx.isel(time=0)
    .to_array("band")
    .plot.imshow(
        col="band",
        size=4,
        vmin=0,
        vmax=4000,
    )
)

# %% [markdown]
# ## Load with bounding box
#
# As you can see `stac_load` returned all the data covered by STAC items
# returned from the query. This happens by default as `stac_load` has no way of
# knowing what your query was. But it is possible to control what region is
# loaded. There are several mechanisms available, but probably simplest one is
# to use `bbox=` parameter (compatible with `stac_client`).
#
# Let's load a small region at native resolution to demonstrate.

# %%
r = 6.5 * km2deg
small_bbox = (x - r, y - r, x + r, y + r)

yy = stac_load(
    items,
    bands=("red", "green", "blue"),
    crs=crs,
    resolution=10,
    chunks={},  # <-- use Dask
    groupby="solar_day",
    bbox=small_bbox,
)
display(yy.odc.geobox)

# %%
yy = yy.compute()

# %%
_ = (
    yy.isel(time=0)
    .to_array("band")
    .plot.imshow(
        col="band",
        size=4,
        vmin=0,
        vmax=4000,
    )
)

# %% [markdown]
# --------------------------------------------------------------



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.13.8
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Access Sentinel 2 Data on Planetary Computer
#
# [![Binder](https://mybinder.org/badge_logo.svg)](https://mybinder.org/v2/gh/opendatacube/odc-stac/develop?labpath=notebooks%2Fstac-load-S2-ms.ipynb)

# %% [markdown]
# ## Setup Instructions
#
# This notebook is meant to run on Planetary Computer lab hub.

# %%
import dask.distributed
import dask.utils
import numpy as np
import planetary_computer as pc
import xarray as xr
from IPython.display import display
from pystac_client import Client

from odc.stac import configure_rio, stac_load

# %% [markdown]
# ## Start Dask Client
#
# This step is optional, but it does improve load speed significantly. You
# don't have to use Dask, as you can load data directly into memory of the
# notebook.

# %%
client = dask.distributed.Client()
configure_rio(cloud_defaults=True, client=client)
display(client)

# %% [markdown]
# ## Query STAC API
#
# Here we are looking for datasets in `sentinel-2-l2a` collection from June
# 2019 over MGRS tile `06VVN`.

# %%
catalog = Client.open("https://planetarycomputer.microsoft.com/api/stac/v1")

query = catalog.search(
    collections=["sentinel-2-l2a"],
    datetime="2019-06",
    query={"s2:mgrs_tile": dict(eq="06VVN")},
)

items = list(query.items())
print(f"Found: {len(items):d} datasets")

# %% [markdown]
# ## Lazy load all the bands
#
# We won't use all the bands but it doesn't matter as bands that we won't use
# won't be loaded. We are "loading" data with Dask, which means that at this
# point no reads will be happening just yet.
#
# We have to supply `dtype=` and `nodata=` because items in this collection are missing [raster extension](https://github.com/stac-extensions/raster) metadata.

# %%
resolution = 10
SHRINK = 4
if client.cluster.workers[0].memory_manager.memory_limit < dask.utils.parse_bytes("4G"):
    SHRINK = 8  # running on Binder with 2Gb RAM

if SHRINK > 1:
    resolution = resolution * SHRINK

xx = stac_load(
    items,
    chunks={"x": 2048, "y": 2048},
    patch_url=pc.sign,
    resolution=resolution,
    # force dtype and nodata
    dtype="uint16",
    nodata=0,
)

print(f"Bands: {','.join(list(xx.data_vars))}")
display(xx)

# %% [markdown]
# By default `stac_load` will return all the data bands using canonical asset
# names. But we can also request a subset of bands, by supplying `bands=` parameter.
# When going this route you can also use "common name" to refer to a band.
#
# In this case we request `red,green,blue,nir` bands which are common names for
# bands `B04,B03,B02,B08` and `SCL` band which is a canonical name.

# %%
xx = stac_load(
    items,
    bands=["red", "green", "blue", "nir", "SCL"],
    resolution=resolution,
    chunks={"x": 2048, "y": 2048},
    patch_url=pc.sign,
    # force dtype and nodata
    dtype="uint16",
    nodata=0,
)

print(f"Bands: {','.join(list(xx.data_vars))}")
display(xx)


# %% [markdown]
# ## Do some math with bands


# %%
def to_float(xx):
    _xx = xx.astype("float32")
    nodata = _xx.attrs.pop("nodata", None)
    if nodata is None:
        return _xx
    return _xx.where(xx != nodata)


def colorize(xx, colormap):
    return xr.DataArray(colormap[xx.data], coords=xx.coords, dims=(*xx.dims, "band"))


# %%
# like .astype(float32) but taking care of nodata->NaN mapping
nir = to_float(xx.nir)
red = to_float(xx.red)
ndvi = (nir - red) / (
    nir + red
)  # < This is still a lazy Dask computation (no data loaded yet)

# Get the 5-th time slice `load->compute->plot`
_ = ndvi.isel(time=4).compute().plot.imshow(size=7, aspect=1.2, interpolation="bicubic")

# %% [markdown]
# For sample purposes work with first 6 observations only

# %%
xx = xx.isel(time=np.s_[:6])

# %%
# fmt: off
scl_colormap = np.array(
    [
        [255,   0, 255, 255],  # 0  - NODATA
        [255,   0,   4, 255],  # 1  - Saturated or Defective
        [0  ,   0,   0, 255],  # 2  - Dark Areas
        [97 ,  97,  97, 255],  # 3  - Cloud Shadow
        [3  , 139,  80, 255],  # 4  - Vegetation
        [192, 132,  12, 255],  # 5  - Bare Ground
        [21 , 103, 141, 255],  # 6  - Water
        [117,   0,  27, 255],  # 7  - Unclassified
        [208, 208, 208, 255],  # 8  - Cloud
        [244, 244, 244, 255],  # 9  - Definitely Cloud
        [195, 231, 240, 255],  # 10 - Thin Cloud
        [222, 157, 204, 255],  # 11 - Snow or Ice
    ],
    dtype="uint8",
)
# fmt: on

# Load SCL band, then convert to RGB using color scheme above
scl_rgba = colorize(xx.SCL.compute(), scl_colormap)

# Check we still have geo-registration
scl_rgba.odc.geobox

# %%
_ = scl_rgba.plot.imshow(col="time", col_wrap=3, size=3, interpolation="antialiased")

# %% [markdown]
# Let's save image dated 2019-06-04 to a cloud optimized geotiff file.

# %%
to_save = scl_rgba.isel(time=3)
fname = f"SCL-{to_save.time.dt.strftime('%Y%m%d').item()}.tif"
print(f"Saving to: '{fname}'")

# %%
scl_rgba.isel(time=3).odc.write_cog(
    fname,
    overwrite=True,
    compress="webp",
    webp_quality=90,
)

# %% [markdown]
# Check the file with `rio info`.

# %%
# !ls -lh {fname}
# !rio info {fname} | jq .

# %% [markdown]
# --------------------------------



"""
STAC -> EO3 utilities.

Utilities for translating STAC Items to EO3 Datasets.
"""

# pylint: disable=too-many-lines
from __future__ import annotations

import datetime
import functools
from collections import Counter
from copy import copy
from typing import (
    Any,
    Dict,
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    TypeAlias,
    Union,
)

import numpy as np
import pystac.asset
import pystac.collection
import pystac.errors
import pystac.item
import shapely.geometry
import xarray as xr
from affine import Affine
from odc.geo import (
    CRS,
    XY,
    Geometry,
    MaybeCRS,
    Resolution,
    SomeResolution,
    geom,
    res_,
    wh_,
    xy_,
)
from odc.geo.geobox import AnchorEnum, GeoBox, GeoboxAnchor
from odc.geo.types import Unset
from odc.geo.xr import ODCExtension
from odc.loader import ReaderDriver, RioDriver, reader_driver
from odc.loader.types import (
    AuxBandMetadata,
    AuxDataSource,
    AuxLoadParams,
    BandKey,
    BandQuery,
    GlobalLoadContext,
    MDParser,
    RasterBandMetadata,
    RasterGroupMetadata,
    RasterSource,
    ReaderDriverSpec,
    norm_nodata,
    with_default,
)
from pystac.extensions.eo import EOExtension
from pystac.extensions.item_assets import ItemAssetsExtension
from pystac.extensions.projection import ProjectionExtension
from pystac.extensions.raster import RasterBand, RasterExtension
from toolz import dicttoolz

from .model import (
    MDParseConfig,
    ParsedItem,
    PropertyLoadRequest,
    RasterCollectionMetadata,
)

ConversionConfig: TypeAlias = Dict[str, Any]

EPSG4326 = CRS("EPSG:4326")

# Assets with these roles are ignored unless manually requested
ROLES_THUMBNAIL = {"thumbnail", "overview"}

# Used to detect image assets when media_type is missing
RASTER_FILE_EXTENSIONS = {
    "tif",
    "tiff",
    "jpeg",
    "jpg",
    "jp2",
    "img",
    "hdf",
    "nc",
    "zarr",
}

# image/* and these media-type are considered to be raster
NON_IMAGE_RASTER_MEDIA_TYPES = {
    "application/x-hdf",
    "application/x-hdf5",
    "application/hdf",
    "application/hdf5",
    "application/x-netcdf",
    "application/netcdf",
    "application/x-zarr",
    "application/zarr",
}


def _band_metadata_raw(asset: pystac.asset.Asset) -> List[RasterBand]:
    bands = asset.to_dict().get("raster:bands", None)
    if bands is None:
        return []
    return [RasterBand(props) for props in bands]


def band_metadata(
    asset: pystac.asset.Asset, default: RasterBandMetadata
) -> List[RasterBandMetadata]:
    """
    Compute band metadata from Asset raster extension with defaults from default.

    :param asset: Asset with raster extension
    :param default: Values to use for fallback
    :return: List of BandMetadata constructed from raster:bands metadata
    """
    bands: List[RasterBand] = []
    try:
        rext = RasterExtension.ext(asset)
        if rext.bands is not None:
            bands = rext.bands
    except pystac.errors.ExtensionNotImplemented:
        bands = _band_metadata_raw(asset)

    if len(bands) == 0:
        return [default]

    return [
        RasterBandMetadata(
            with_default(band.data_type, default.data_type),
            with_default(norm_nodata(band.nodata), default.nodata),
            with_default(band.unit, default.units),
        )
        for band in bands
    ]


def has_proj_ext(item: Union[pystac.item.Item, pystac.collection.Collection]) -> bool:
    """
    Check if STAC Item or Collection has projection extension.

    :returns: ``True`` if PROJ extension is enabled
    :returns: ``False`` if no PROJ extension was found
    """
    if ProjectionExtension.has_extension(item):
        return True
    # can remove this block once pystac 1.9.0 is the min supported version
    return any(
        ext_name.startswith("https://stac-extensions.github.io/projection/")
        for ext_name in item.stac_extensions
    )


def has_raster_ext(item: Union[pystac.item.Item, pystac.collection.Collection]) -> bool:
    """
    Check if STAC Item/Collection have RasterExtension.

    :returns: ``True`` if Raster extension is enabled
    :returns: ``False`` if no Raster extension was found
    """
    if RasterExtension.has_extension(item):
        return True
    # can remove this block once pystac 1.9.0 is the min supported version
    return any(
        ext_name.startswith("https://stac-extensions.github.io/raster/")
        for ext_name in item.stac_extensions
    )


def has_proj_data(asset: pystac.asset.Asset) -> bool:
    """
    Check if STAC Asset contains proj extension data.

    :returns: True if both ``.shape`` and ``.transform`` are set
    :returns: False if either ``.shape`` or ``.transform`` are missing
    """
    prj = ProjectionExtension.ext(asset)
    return prj.shape is not None and prj.transform is not None


def is_raster_data(asset: pystac.asset.Asset, check_proj: bool = False) -> bool:
    """
    Heuristic for determining if Asset points to raster data.

    - If media type looks like image and roles don't look like thumbnail/overview
    - If media type is undefined and roles contains "data"
    - If media type is undefined and href ends on image extension

    :param asset:
       STAC Asset to check

    :param check_proj:
       when enabled check if asset is part of an Item that has projection
       extension enabled and if yes only consider bands with
       projection data as "raster data" bands.
    """
    # pylint: disable=too-many-return-statements
    #   some of these are for readability

    if check_proj:
        if (
            asset.owner is not None
            and has_proj_ext(asset.owner)  # type: ignore
            and not has_proj_data(asset)
        ):
            return False

    roles: Set[str] = set(asset.roles or [])

    media_type = asset.media_type
    if media_type is None:
        # Type undefined
        #   Look if it has data role
        if "data" in roles:
            return True
        if "metadata" in roles:
            return False

        ext = asset.href.split(".")[-1].lower()
        return ext in RASTER_FILE_EXTENSIONS

    media_type, *_ = media_type.split(";")
    media_type = media_type.lower()

    if media_type.startswith("image/"):
        # Image:
        #    False -- when thumbnail
        #    True  -- otherwise
        if any(r in roles for r in ROLES_THUMBNAIL):
            return False
        return True

    if media_type in NON_IMAGE_RASTER_MEDIA_TYPES:
        return True

    # some unsupported mime type
    return False


def mk_1x1_geobox(g: Geometry) -> GeoBox:
    """
    Construct 1x1 pixels GeoBox tightly enclosing supplied geometry.

    :param g: Geometry in whatever projection
    :return: GeoBox object such that geobox.extent.contains(g) is True, geobox.shape == (1,1)
    """
    x1, y1, x2, y2 = g.boundingbox
    # note that Y axis is inverted
    #   0,0 -> X_min, Y_max
    #   1,1 -> X_max, Y_min
    return GeoBox((1, 1), Affine((x2 - x1), 0, x1, 0, (y1 - y2), y2), g.crs)


def asset_geobox(asset: pystac.asset.Asset) -> GeoBox:
    """
    Compute GeoBox from STAC Asset.

    This only works if ProjectionExtension is used with the
    following properties populated:

    - shape
    - transform
    - CRS

    :raises ValueError: when transform,shape or crs are missing
    :raises ValueError: when transform is not Affine.
    """
    try:
        _proj = ProjectionExtension.ext(asset)
    except pystac.errors.ExtensionNotImplemented:
        raise ValueError("No projection extension defined") from None

    if _proj.shape is None or _proj.transform is None or _proj.crs_string is None:
        raise ValueError(
            "The asset must have the following fields (from the projection extension):"
            " shape, transform, and one of an epsg, wkt2, or projjson"
        )

    h, w = _proj.shape
    if len(_proj.transform) not in (6, 9):
        raise ValueError("Asset transform must be 6 or 9 elements in size")

    if len(_proj.transform) == 9 and _proj.transform[-3:] != [0, 0, 1]:
        raise ValueError(f"Asset transform is not affine: {_proj.transform}")

    affine = Affine(*_proj.transform[:6])
    return GeoBox(wh_(w, h), affine, _proj.crs_string)


def geobox_gsd(geobox: GeoBox) -> float:
    """
    Compute ground sampling distance of a given GeoBox.

    :param geobox: input :class:`~odc.geo.geobox.GeoBox`
    :returns: Minimum ground sampling distance along X/Y
    """
    return min(map(abs, [geobox.transform.a, geobox.transform.e]))


def compute_eo3_grids(
    assets: Dict[str, pystac.asset.Asset],
) -> Tuple[Dict[str, GeoBox], Dict[str, str]]:
    """
    Compute a minimal set of eo3 grids.

    Pick default one, give names to non-default grids, while keeping track of
    which asset has which grid

    Assets must have ProjectionExtension with shape, transform and crs information
    populated.
    """
    geoboxes = dicttoolz.valmap(asset_geobox, assets)
    return _group_geoboxes(geoboxes)


def _group_geoboxes(
    geoboxes: Dict[str, GeoBox],
) -> Tuple[Dict[str, GeoBox], Dict[str, str]]:
    # pylint: disable=too-many-locals
    if len(geoboxes) == 0:
        return {}, {}

    def gbox_name(geobox: GeoBox) -> str:
        gsd = geobox_gsd(geobox)
        return f"g{gsd:g}"

    # GeoBox to list of bands that share same footprint
    grids: Dict[GeoBox, List[str]] = {}
    crs: Optional[CRS] = None

    for k, geobox in geoboxes.items():
        grids.setdefault(geobox, []).append(k)

    # Default grid is the one with highest count of bands
    #   If there is a tie pick one with the smallest ground sampling distance
    def gbox_score(geobox: GeoBox) -> Tuple[int, float]:
        return (-len(grids[geobox]), geobox_gsd(geobox))

    # locate default grid
    g_default, *_ = sorted(grids, key=gbox_score)

    named_grids: Dict[str, GeoBox] = {}
    band2grid: Dict[str, str] = {}
    for grid, bands in grids.items():
        if crs is None:
            crs = grid.crs

        grid_name = "default" if grid is g_default else gbox_name(grid)
        if grid_name in named_grids:
            band, *_ = bands
            grid_name = f"{grid_name}-{band}"

        named_grids[grid_name] = grid
        for band in bands:
            band2grid[band] = grid_name

    return named_grids, band2grid


def band2grid_from_gsd(assets: Dict[str, pystac.asset.Asset]) -> Dict[str, str]:
    if not assets:
        return {}

    grids: Dict[float, List[str]] = {}
    for name, asset in assets.items():
        gsd = asset.common_metadata.gsd
        gsd = 0 if gsd is None else gsd
        gsd_normed = float(f"{gsd:g}")
        grids.setdefault(gsd_normed, []).append(name)

    # Default grid is one with largest number of bands
    # .. and lowest gsd when there is a tie
    (_, default_gsd), *_ = sorted((-len(bands), gsd) for gsd, bands in grids.items())
    band2grid = {}
    for gsd, bands in grids.items():
        grid_name = "default" if gsd == default_gsd else f"g{gsd:g}"
        for band in bands:
            band2grid[band] = grid_name

    return band2grid


def _extract_aliases(
    asset_name: str,
    asset: pystac.asset.Asset,
    block_list: Set[str],
) -> Iterator[Tuple[str, int, BandKey]]:
    try:
        eo = EOExtension.ext(asset)
    except pystac.errors.ExtensionNotImplemented:
        return
    if eo.bands is None:
        return

    for idx, band in enumerate(eo.bands):
        for alias in [band.name, band.common_name]:
            if alias is not None and alias not in block_list:
                yield (alias, len(eo.bands), (asset_name, idx + 1))


def alias_map_from_eo(item: pystac.item.Item) -> Dict[str, List[BandKey]]:
    """
    Generate mapping ``common name -> canonical name``.

    For all unique common names defined on the Item's assets via the eo
    extension, record a mapping to the asset key ("canonical name"). Non-unique
    common names are ignored with a warning unless ``quiet`` flag is set.

    :param item: STAC :class:`~pystac.item.Item` to process
    :return: common name to (asset, idx) mapping
    """
    aliases: Dict[str, List[BandKey]] = {}

    asset_band_counts: Dict[str, int] = {}
    asset_names = set(item.assets)
    for asset_name, asset in item.assets.items():
        for alias, count, bkey in _extract_aliases(asset_name, asset, asset_names):
            aliases.setdefault(alias, []).append(bkey)
            asset_band_counts[asset_name] = count

    # Alias pointing to an asset with fewer bands is
    # of higher priority, 1-band data asset vs 3 band visual
    def _cmp(x):
        asset, _ = x
        return (asset_band_counts[asset], asset)

    return {alias: sorted(bands, key=_cmp) for alias, bands in aliases.items()}


def mk_sample_item(collection: pystac.collection.Collection) -> pystac.item.Item:
    item_assets = getattr(collection, "item_assets", None)
    if item_assets is None:
        try:
            item_assets = ItemAssetsExtension.ext(collection).item_assets
        except pystac.errors.ExtensionNotImplemented:
            pass

    if not item_assets:
        raise ValueError("This only works on Collections with ItemAssets extension")

    item = pystac.item.Item(
        "sample",
        None,
        None,
        datetime.datetime(2020, 1, 1),
        {},
        stac_extensions=collection.stac_extensions,
        collection=collection,
    )

    for name, asset in item_assets.items():
        _asset = {"href": ""}
        _asset.update(asset.to_dict())
        item.add_asset(name, pystac.asset.Asset.from_dict(_asset))

    return item


def _collection_id(item: pystac.item.Item) -> str:
    # choose first that is set
    # 1. collection_id
    # 2. odc:product
    # 3. "_"
    if item.collection_id is None:
        # early ODC data
        return str(item.properties.get("odc:product", "_"))
    return str(item.collection_id)


class StacMDParser:
    """
    Extract metadata from STAC items.

    implements :class:`~odc.loader.types.MDParser` protocol.
    """

    def __init__(self, cfg: ConversionConfig):
        self._cfg = cfg

    @functools.lru_cache()
    def _config(self, collection_id: str | None) -> MDParseConfig:
        return MDParseConfig.from_dict(self._cfg, collection_id)

    def extract(self, md: Any) -> RasterGroupMetadata:
        # pylint: disable=too-many-locals
        assert isinstance(md, pystac.item.Item)
        item = md
        c = self._config(item.collection_id)

        def _keep(kv: tuple[str, pystac.asset.Asset], check_proj: bool) -> bool:
            name, asset = kv
            if name in c.band_cfg:
                return True
            return is_raster_data(asset, check_proj=check_proj)

        # Ignore non-proj data bands when Item has proj extension, unless user
        # disabled that filter with `ignore_proj=True` option
        check_proj = (
            has_proj_ext(item)
            and not c.ignore_proj
            and not any(has_proj_data(a) for a in item.assets.values())
        )
        data_bands = dicttoolz.itemfilter(lambda kv: _keep(kv, check_proj), item.assets)

        if len(data_bands) == 0 and check_proj:
            # If no data bands found with check_proj=True, fallback to check_proj=False
            # This handles items that declare proj extension at item level but don't have
            # per-asset proj data (shape/transform)
            data_bands = dicttoolz.itemfilter(lambda kv: _keep(kv, False), item.assets)

        bands: dict[BandKey, RasterBandMetadata | AuxBandMetadata] = {}
        aliases = alias_map_from_eo(item)

        # 1. If band in user config -- use that
        # 2. Use data from raster extension (with fallback to "*" config)
        # 3. Use config for "*" from user config as fallback
        for name, asset in data_bands.items():
            bands.update(self._extract_bands(name, asset, c))

        for alias, bkey in c.aliases.items():
            aliases.setdefault(alias, []).insert(0, bkey)

        for idx, prop in enumerate(c.with_props):
            bk: BandKey = ("_stac_metadata", idx + 1)
            bands[bk] = AuxBandMetadata(
                prop.dtype,
                nodata=prop.nodata,
                units=prop.units,
                driver_data=prop,
            )
            aliases[prop.output_name] = [bk]

        return RasterGroupMetadata(bands, aliases, c.extra_dims, c.extra_coords)

    def driver_data(self, md: Any, band_key: BandKey) -> Any:
        # None for raster bands
        # (PropertyLoadRequest, Value|None)
        assert isinstance(md, pystac.item.Item)
        c = self._config(md.collection_id)

        asset_name, band_idx = band_key
        driver_data = None

        if asset_name == "_stac_metadata":
            prop_cfg = c.with_props[band_idx - 1]
            driver_data = md.properties.get(prop_cfg.key, None)

        return driver_data

    def _extract_bands(
        self, name: str, asset: pystac.asset.Asset, cfg: MDParseConfig
    ) -> dict[BandKey, RasterBandMetadata]:
        bm = cfg.band_cfg.get(name, None)
        if bm is not None:
            return {(name, 1): copy(bm)}

        bm = cfg.band_cfg.get(f"{name}.*", None)
        if bm is None:
            bm = cfg.band_defaults

        bands = band_metadata(asset, bm)

        return {(name, idx + 1): bm for idx, bm in enumerate(bands)}


class StacAuxReader:
    """
    Implements AuxReader protocol for STAC items.

    Handles reading auxiliary data from STAC items, particularly metadata properties
    that are exposed as auxiliary bands.
    """

    # pylint: disable=too-few-public-methods

    def read(
        self,
        srcs: Sequence[Sequence[AuxDataSource]],
        cfg: AuxLoadParams,
        used_names: set[str],
        available_coords: Mapping[str, xr.DataArray],
        ctx: GlobalLoadContext,
        *,
        dask_layer_name: str | None = None,
    ) -> xr.DataArray:
        """
        Read auxiliary data from STAC items.

        :param srcs: Auxiliary data sources grouped by time
        :param cfg: Loading configuration
        :param used_names: Names claimed by raster bands and their coordinates
        :param available_coords: Available coordinates, must include time
        :param ctx: Load context
        :param dask_layer_name: Suggested dask layer name when reading with dask
        :return: Auxiliary data loaded into a xarray.DataArray
        """
        assert (used_names, ctx, dask_layer_name) is not None
        # cfg.meta.driver_data: PropertyLoadRequest
        # srcs[].driver_data:  None|float|str|int

        def extract_cfg() -> PropertyLoadRequest:
            meta = cfg.meta
            assert meta is not None
            assert isinstance(meta.driver_data, PropertyLoadRequest)
            return meta.driver_data

        prop_cfg = extract_cfg()
        _fill = prop_cfg.fill_value
        assert _fill is not None

        def _value(row: Sequence[AuxDataSource]) -> Any:
            if len(row) == 0:
                return _fill

            dd: Iterator[Any] = (src.driver_data for src in row)
            return prop_cfg.fuser([v for v in dd if v is not None])

        values = [_value(row) for row in srcs]
        data = np.array(values, dtype=cfg.dtype)
        attrs: dict[str, Any] = {"units": prop_cfg.units}
        if cfg.fill_value is not None:
            attrs["nodata"] = cfg.fill_value

        # Use time coordinate
        time = available_coords["time"]
        return xr.DataArray(
            data,
            coords={"time": time},
            dims=["time"],
            attrs=attrs,
        )


class _CMDAssembler:
    """
    Incrementally build up collection metadata from item stream.

    Expect to see items of the same collection only.
    """

    # pylint: disable=too-few-public-methods,too-many-instance-attributes

    def __init__(
        self,
        md_plugin: MDParser,
        collection_id: str,
    ) -> None:
        self.collection_id = collection_id
        self.md: Optional[RasterCollectionMetadata] = None
        self.md_plugin = md_plugin
        self._asset_keeps: dict[str, bool] = {}
        self._known_assets: set[str] = set()

    def _bootstrap(self, item: pystac.item.Item) -> None:
        """Called on the very first item only."""
        assert self.md_plugin is not None

        has_proj = has_proj_ext(item)
        meta = self.md_plugin.extract(item)
        data_asset_names = set(n for n, _ in meta.bands if n in item.assets)
        data_assets = {n: item.assets[n] for n in data_asset_names}

        # We assume that grouping of data bands into grids is consistent across
        # the entire collection, so we compute it once and keep it
        if has_proj and data_assets:
            # Check if any data assets have proj data.
            if any(has_proj_data(a) for a in data_assets.values()):
                _, band2grid = compute_eo3_grids(data_assets)
            else:
                band2grid = band2grid_from_gsd(data_assets)
                has_proj = False
        else:
            band2grid = band2grid_from_gsd(data_assets)

        self._asset_keeps = {name: name in data_assets for name in item.assets}
        self._known_assets = set(self._asset_keeps)

        self.md = RasterCollectionMetadata(
            self.collection_id,
            meta,
            has_proj=has_proj,
            band2grid=band2grid,
        )

    def update(self, item: pystac.item.Item) -> None:
        # pylint: disable=too-many-locals,too-many-branches
        if self.md is None:
            self._bootstrap(item)
            return

        meta = self.md.meta.merge(self.md_plugin.extract(item))
        self.md = self.md.patch(meta=meta)

        new_assets = set(item.assets) - self._known_assets
        if len(new_assets) == 0:
            return

        has_proj = has_proj_ext(item)
        new_data_assets: list[tuple[str, pystac.asset.Asset]] = []
        data_asset_names = set(n for n, _ in meta.bands)
        for name in new_assets:
            asset = item.assets[name]
            is_data = name in data_asset_names
            self._asset_keeps[name] = is_data
            if is_data:
                new_data_assets.append((name, asset))
        self._known_assets = set(self._asset_keeps)

        # no new assets we care about
        if len(new_data_assets) == 0:
            return

        if not has_proj:
            return

        # Update band2grid
        band2grid = self.md.band2grid

        # GeoBox -> grid name
        grid2band: dict[GeoBox, str] = {}
        for name, asset in item.assets.items():
            if (grid_name := band2grid.get(name, None)) is not None:
                grid2band[asset_geobox(asset)] = grid_name

        for name, asset in new_data_assets:
            band2grid[name] = grid2band.get(asset_geobox(asset), f"grid-{name}")


def extract_collection_metadata(
    item: pystac.item.Item,
    cfg: Optional[ConversionConfig] = None,
    md_plugin: MDParser | None = None,
) -> RasterCollectionMetadata:
    """
    Use sample item to figure out raster bands within the collection.

    1. Decide which assets contain raster data
    2. Extract metadata about rasters from STAC or from ``cfg``
    3. See if ``proj`` data is available and group bands by resolution
    4. Construct alias map from common names and user config

    :param item: Representative STAC item from a collection.
    :param cfg: Optional user configuration
    :return: :py:class:`~odc.stac._model.RasterCollectionMetadata`
    """
    _, md_plugin = _resolve_driver(None, cfg)
    proc = _CMDAssembler(md_plugin, _collection_id(item))
    proc.update(item)
    assert proc.md is not None
    return proc.md


def parse_item(
    item: pystac.item.Item,
    template: ConversionConfig | RasterCollectionMetadata | None = None,
    md_plugin: MDParser | None = None,
    asset_absolute_paths: bool = True,
) -> ParsedItem:
    """
    Extract raster band information relevant for data loading.

    :param item: STAC Item
    :param cfg: Common collection level information
    :return: ``ParsedItem``
    """

    if template is None or isinstance(template, dict):
        return next(parse_items([item], template, md_plugin))

    # TODO: remove this part, i.e. template = RasterCollectionMetadata(...)
    # version of this method
    assert isinstance(template, RasterCollectionMetadata)
    if md_plugin is None:
        _, md_plugin = _resolve_driver(None, None)

    return _parse_item(item, template, md_plugin, asset_absolute_paths)


def _parse_item(
    item: pystac.item.Item,
    template: RasterCollectionMetadata,
    md_plugin: MDParser,
    asset_absolute_paths: bool = True,
) -> ParsedItem:
    """
    Extract raster band information relevant for data loading.

    :param item: STAC Item
    :param template: Common collection level information
    :param asset_absolute_paths: Use absolute paths for assets
    :return: ``ParsedItem``
    """
    # pylint: disable=too-many-locals
    band2grid = template.band2grid
    has_proj = False if template.has_proj is False else has_proj_ext(item)
    _assets = item.assets

    _grids: Dict[str, GeoBox] = {}
    bands: Dict[BandKey, RasterSource | AuxDataSource] = {}
    geometry: Optional[Geometry] = None

    if item.geometry is not None:
        geometry = Geometry(item.geometry, EPSG4326)

    def _get_grid(grid_name: str, asset: pystac.asset.Asset) -> GeoBox:
        grid = _grids.get(grid_name, None)
        if grid is not None:
            return grid
        grid = asset_geobox(asset)
        _grids[grid_name] = grid
        return grid

    def _get_driver_data(bk: BandKey) -> tuple[Any, str | None]:
        driver_data: Any = None
        subdataset: str | None = None
        driver_data = md_plugin.driver_data(item, bk)
        if isinstance(driver_data, dict):
            subdataset = driver_data.get("subdataset", None)
        return driver_data, subdataset

    for bk, meta in template.meta.bands.items():
        uri: str | None = None
        asset_name, band_idx = bk
        asset = _assets.get(asset_name)

        if asset is not None:
            uri = asset.get_absolute_href() if asset_absolute_paths else asset.href
            if uri is None:
                raise ValueError(
                    f"Can not determine absolute path for asset: {asset_name}"
                )  # pragma: no cover (https://github.com/stac-utils/pystac/issues/754)

        if isinstance(meta, RasterBandMetadata):
            if asset is None or uri is None:
                continue

            driver_data, subdataset = _get_driver_data(bk)
            grid_name = band2grid.get(asset_name, "default")
            geobox: Optional[GeoBox] = _get_grid(grid_name, asset) if has_proj else None

            # Assumption: if extra dims are defined then asset bands are loaded into 3d+ array
            # RasterSource.band == 0 indicates "all the bands"
            if meta.extra_dims:
                band_idx = 0

            bands[bk] = RasterSource(
                uri=uri,
                band=band_idx,
                subdataset=subdataset,
                geobox=geobox,
                meta=meta,
                driver_data=driver_data,
            )
        elif isinstance(meta, AuxBandMetadata):
            if uri is None:
                uri = f"virtual://{asset_name}/{band_idx}"

            driver_data, subdataset = _get_driver_data(bk)
            bands[bk] = AuxDataSource(
                uri=uri,
                subdataset=subdataset,
                meta=meta,
                driver_data=driver_data,
            )

    data_asset_names = set(template.asset_names())
    accessories = {
        name: asset.to_dict()
        for name, asset in _assets.items()
        if name not in data_asset_names
    }

    _cmd = item.common_metadata
    return ParsedItem(
        item.id,
        template,
        bands,
        geometry,
        datetime=item.datetime,
        datetime_range=(_cmd.start_datetime, _cmd.end_datetime),
        href=item.get_self_href(),
        accessories=accessories,
    )


def parse_items(
    items: Iterable[pystac.item.Item],
    cfg: ConversionConfig | None = None,
    md_plugin: MDParser | None = None,
    asset_absolute_paths: bool = True,
) -> Iterator[ParsedItem]:
    """
    Parse sequence of STAC Items into internal representation.

    Exposed for debugging purposes.
    """
    proc_cache: Dict[str, _CMDAssembler] = {}

    for item in items:
        collection_id = _collection_id(item)
        proc = proc_cache.get(collection_id, None)
        if proc is None:
            if md_plugin is None:
                _, md_plugin = _resolve_driver(None, cfg)
            proc = _CMDAssembler(md_plugin, collection_id)
            proc_cache[collection_id] = proc

        proc.update(item)
        assert proc.md is not None
        yield parse_item(item, proc.md, proc.md_plugin, asset_absolute_paths)


def _most_common_gbox(
    gboxes: Sequence[GeoBox],
    thresh: float = 0.1,
) -> Tuple[Optional[CRS], Resolution, GeoboxAnchor, Optional[GeoBox]]:
    gboxes = list(gboxes)

    # First check for identical geoboxes
    _gboxes = set(gboxes)
    if len(_gboxes) == 1:
        g = _gboxes.pop()
        return (g.crs, g.resolution, g.anchor, g)

    # Most common shared CRS, Resolution, Anchor
    gg = [(g.crs, g.resolution, g.anchor) for g in gboxes]
    hist = Counter(gg)
    (best, n), *_ = hist.most_common(1)
    if n / len(gg) > thresh:
        return (*best, None)

    # too few in the majority group
    # redo ignoring anchor this time
    hist = Counter((crs, res) for crs, res, _ in gg)
    (best, _), *_ = hist.most_common(1)
    return (*best, AnchorEnum.EDGE, None)


def _auto_load_params(
    items: Sequence[ParsedItem], bands: Optional[Sequence[str]] = None
) -> Optional[Tuple[Optional[CRS], Resolution, GeoboxAnchor, Optional[GeoBox]]]:
    def _extract_gbox(
        item: ParsedItem,
    ) -> Optional[GeoBox]:
        gbx = item.geoboxes(bands)
        return gbx[0] if len(gbx) else None

    geoboxes = [gbox for gbox in map(_extract_gbox, items) if gbox is not None]
    if len(geoboxes) == 0:
        return None

    return _most_common_gbox(geoboxes, 0.1)


def _normalize_geometry(xx: Any) -> Geometry:
    if isinstance(xx, shapely.geometry.base.BaseGeometry):
        return Geometry(xx, "epsg:4326")

    if isinstance(xx, Geometry):
        return xx

    if isinstance(xx, dict):
        return Geometry(xx, "epsg:4326")

    # GeoPandas
    _geo = getattr(xx, "__geo_interface__", None)
    if _geo is None:
        raise ValueError("Can't interpret value as geometry")

    _crs = getattr(xx, "crs", "epsg:4326")
    return Geometry(_geo, _crs)


def _compute_bbox(
    items: Iterable[ParsedItem],
    crs: MaybeCRS,
    bands: BandQuery = None,
) -> geom.BoundingBox:
    def bboxes(items: Iterable[ParsedItem]) -> Iterator[geom.BoundingBox]:
        crs0 = crs
        for item in items:
            g = item.safe_geometry(crs0, bands=bands)
            assert g is not None
            if crs0 is crs:
                # If crs is something like "utm", make sure
                # same one is used going forward
                crs0 = g.crs
            yield g.boundingbox

    return geom.bbox_union(bboxes(items))


def _align2anchor(
    align: Optional[Union[float, int, XY[float]]], resolution: SomeResolution
) -> GeoboxAnchor:
    if align is None:
        return AnchorEnum.EDGE

    if isinstance(align, (float, int)):
        align = xy_(align, align)

    # support old-style "align", which is basically anchor but in CRS units
    ax, ay = align.xy
    if ax == 0 and ay == 0:
        return AnchorEnum.EDGE
    resolution = res_(resolution)
    return xy_(ax / abs(resolution.x), ay / abs(resolution.y))


def output_geobox(
    items: Sequence[ParsedItem],
    bands: Optional[Sequence[str]] = None,
    *,
    crs: MaybeCRS = Unset(),
    resolution: Optional[SomeResolution] = None,
    anchor: Optional[GeoboxAnchor] = None,
    align: Optional[Union[float, int, XY[float]]] = None,
    geobox: Optional[GeoBox] = None,
    like: Optional[Any] = None,
    geopolygon: Optional[Any] = None,
    bbox: Optional[Tuple[float, float, float, float]] = None,
    lon: Optional[Tuple[float, float]] = None,
    lat: Optional[Tuple[float, float]] = None,
    x: Optional[Tuple[float, float]] = None,
    y: Optional[Tuple[float, float]] = None,
) -> Optional[GeoBox]:
    """
    Used to compute output geobox from load parameters.

    Exposed at top-level for debugging.
    """
    # pylint: disable=too-many-locals,too-many-branches,too-many-statements
    # pylint: disable=too-many-return-statements,too-many-arguments

    # geobox, like --> GeoBox
    # lon,lat      --> geopolygon[epsg:4326]
    # bbox         --> geopolygon[epsg:4326]
    # x,y,crs      --> geopolygon[crs]
    # [items]      --> crs, geopolygon[crs]
    # [items]      --> crs, resolution
    # geopolygon, crs, resolution[, anchor|align] --> GeoBox

    params = {
        k
        for k, v in {
            "x": x,
            "y": y,
            "lon": lon,
            "lat": lat,
            "crs": crs,
            "resolution": resolution,
            "align": align,
            "anchor": anchor,
            "like": like,
            "geopolygon": geopolygon,
            "bbox": bbox,
            "geobox": geobox,
        }.items()
        if not (v is None or isinstance(v, Unset))
    }

    def report_extra_args(primary: str, *ok_args):
        args = params - set([primary, *ok_args])
        if len(args) > 0:
            raise ValueError(
                f"Too many arguments when using `{primary}=`: {','.join(args)}"
            )

    def check_arg_sets(*args: str) -> bool:
        x = params & set(args)
        if len(x) == 0 or len(x) == len(args):
            return True
        return False

    if geobox is not None:
        report_extra_args("geobox")
        return geobox
    if like is not None:
        report_extra_args("like")
        if isinstance(like, GeoBox):
            return like
        _odc = getattr(like, "odc", None)
        if _odc is None:
            raise ValueError("No geospatial info on `like=` input")

        assert isinstance(_odc, ODCExtension)
        if _odc.geobox is None:
            raise ValueError("No geospatial info on `like=` input")

        assert isinstance(_odc.geobox, GeoBox)
        return _odc.geobox

    if not check_arg_sets("x", "y"):
        raise ValueError("Need to supply both x= and y=")

    if not check_arg_sets("lon", "lat"):
        raise ValueError("Need to supply both lon= and lat=")

    if isinstance(crs, Unset):
        crs = None

    grid_params = ("crs", "align", "anchor", "resolution")

    query_crs: Optional[CRS] = None
    if geopolygon is not None:
        geopolygon = _normalize_geometry(geopolygon)
        query_crs = geopolygon.crs

    # Normalize  x.y|lon.lat|bbox|geopolygon arguments to a geopolygon|None
    if geopolygon is not None:
        report_extra_args("geopolygon", *grid_params)
    elif bbox is not None:
        report_extra_args("bbox", *grid_params)
        x0, y0, x1, y1 = bbox
        geopolygon = geom.box(x0, y0, x1, y1, EPSG4326)
    elif lat is not None and lon is not None:
        # lon=(x0, x1), lat=(y0, y1)
        report_extra_args("lon,lat", "lon", "lat", *grid_params)
        x0, x1 = sorted(lon)
        y0, y1 = sorted(lat)
        geopolygon = geom.box(x0, y0, x1, y1, EPSG4326)
    elif x is not None and y is not None:
        if crs is None:
            raise ValueError("Need to supply `crs=` when using `x=`, `y=`.")
        report_extra_args("x,y", "x", "y", *grid_params)
        x0, x1 = sorted(x)
        y0, y1 = sorted(y)
        geopolygon = geom.box(x0, y0, x1, y1, crs)

    full_auto = len(params) == 0
    _anchor: GeoboxAnchor = AnchorEnum.EDGE
    _the_gbox: Optional[GeoBox] = None

    if crs is None or resolution is None:
        rr = _auto_load_params(items, bands)
        if rr is not None:
            _crs, _res, _anchor, _the_gbox = rr
        else:
            _crs, _res = None, None

        if full_auto and _the_gbox is not None:
            return _the_gbox

        if crs is None:
            crs = _crs or query_crs

        if resolution is None:
            resolution = _res

        if resolution is None or crs is None:
            return None

    if anchor is None:
        anchor = _anchor if align is None else _align2anchor(align, resolution)

    if geopolygon is not None:
        assert isinstance(geopolygon, Geometry)
        return GeoBox.from_geopolygon(
            geopolygon,
            resolution=resolution,
            crs=crs,
            anchor=anchor,
        )

    # compute from parsed items
    _bbox = _compute_bbox(items, crs)
    return GeoBox.from_bbox(_bbox, resolution=resolution, anchor=anchor)


def _resolve_driver(
    driver: ReaderDriverSpec | None,
    stac_cfg: Optional[ConversionConfig],
    with_properties: Sequence[str | Mapping[str, Any]] | None = None,
) -> tuple[ReaderDriver, MDParser]:
    md_parser: MDParser | None = None

    stac_cfg = {} if stac_cfg is None else stac_cfg
    if with_properties is not None:
        stac_cfg["with_properties"] = with_properties
    if driver is None:
        md_parser = StacMDParser(stac_cfg)
        return RioDriver(md_parser=md_parser, aux_reader=StacAuxReader()), md_parser
    rdr = reader_driver(driver)
    md_parser = rdr.md_parser
    if md_parser is None:
        md_parser = StacMDParser(stac_cfg)
    return rdr, md_parser



"""Utilities for benchmarking."""

import json
from pathlib import Path
from typing import Any, Dict

# pylint: disable=import-outside-toplevel

SAMPLE_SITES = {
    "s2-ms-mosaic": {
        "file_id": "s2-ms-mosaic_2020-06-06--P1D",
        "api": "https://planetarycomputer.microsoft.com/api/stac/v1",
        "search": {
            "collections": ["sentinel-2-l2a"],
            "datetime": "2020-06-06",
            "bbox": [27.345815, -14.98724, 27.565542, -7.710992],
            "query": {},
        },
    },
    "s2-ms-deep": {
        "file_id": "s2-ms-deep_2020-06--P2M_35MNM",
        "api": "https://planetarycomputer.microsoft.com/api/stac/v1",
        "search": {
            "collections": ["sentinel-2-l2a"],
            "datetime": "2020-06/2020-07",
            "bbox": None,
            "query": {
                "s2:mgrs_tile": {"eq": "35MNM"},
                "s2:nodata_pixel_percentage": {"lt": 10},
            },
        },
    },
}


def dump_site(site: Dict[str, Any], overwrite: bool = False) -> Dict[str, Any]:
    """
    Prepare input for benchmarking.

    Queries API end-point according to site configuration and dumps result into a geojson file. Site
    configuration must include ``file_id:str, api:str, search:Dict[str,Any]``.

    .. code-block:: json

       {
         "file_id": "ms-s2-long-mosaic_2020-06-06--P1D",
         "api": "https://planetarycomputer.microsoft.com/api/stac/v1",
         "search": {
           "collections": ["sentinel-2-l2a"],
           "datetime": "2020-06-06",
           "bbox": [ 27.345815, -14.98724, 27.565542, -7.710992],
           "query": {}
         }
       }

    :param site: Definition of the test query
    :param overwrite: overwrite existing file
    :return: Returns GeoJSON FeatureCollection with extra metadata about the query
    """
    import pystac_client

    api = site["api"]
    search = site["search"]

    cat = pystac_client.Client.open(api)
    search = cat.search(**search)
    print(f"Query API end-point: {api}")
    all_features = search.item_collection_as_dict()
    all_features["properties"] = {
        "api": search.url,
        "search": search._parameters,  # pylint: disable=protected-access
    }

    out_path = Path(f"{site['file_id']}.geojson")
    if out_path.exists():
        if overwrite:
            print(f"Will overwrite: {out_path}")
        else:
            print(f"File exists, keeping previous version: {out_path}")
            return all_features

    print(f"Writing to: {out_path}")
    with open(out_path, "wt", encoding="utf8") as dst:
        json.dump(all_features, dst)

    return all_features



# pylint: disable=missing-module-docstring,redefined-builtin
from unittest.mock import MagicMock

import pystac
import pystac.item
import pytest
import shapely.geometry
from dask.utils import ndeepmap
from odc.geo.geobox import GeoBox
from odc.geo.xr import ODCExtension
from odc.loader import resolve_load_cfg
from odc.loader.types import RasterBandMetadata

from odc.stac import RasterLoadParams
from odc.stac import load as stac_load
from odc.stac._stac_load import _group_items
from odc.stac.testing.stac import b_, mk_parsed_item, to_stac_item


def test_stac_load_smoketest(
    sentinel_stac_ms_with_raster_ext: pystac.item.Item,
) -> None:
    item = sentinel_stac_ms_with_raster_ext.clone()

    params = {"crs": "EPSG:3857", "resolution": 100, "align": 0, "chunks": {}}
    xx = stac_load([item], "B02", **params)  # type: ignore

    assert isinstance(xx.B02.odc, ODCExtension)
    assert xx.B02.shape[0] == 1
    assert xx.B02.odc.geobox is not None
    assert xx.B02.odc.geobox.crs == "EPSG:3857"
    assert xx.time.dtype == "datetime64[ns]"

    # Test dc.load name for bands, and alias support
    xx = stac_load([item], measurements=["red", "green"], **params)  # type: ignore

    assert "red" in xx.data_vars
    assert "green" in xx.data_vars
    assert xx.red.shape == xx.green.shape

    # Test dc.load name for bands, and alias support
    patch_url = MagicMock(return_value="https://example.com/f.tif")
    xx = stac_load(
        [item],
        measurements=["red", "green"],
        patch_url=patch_url,
        **params,  # type: ignore
    )
    assert isinstance(xx.odc, ODCExtension)

    # expect patch_url to be called 2 times, 1 for red and 1 for green band
    assert patch_url.call_count == 2

    patch_url = MagicMock(return_value="https://example.com/f.tif")
    zz = stac_load(
        [item],
        patch_url=patch_url,
        stac_cfg={"*": {"warnings": "ignore"}},
        **params,  # type: ignore
    )
    assert patch_url.call_count == len(zz.data_vars)

    yy = stac_load(
        [item], ["nir"], like=xx, chunks={}, stac_cfg={"*": {"warnings": "ignore"}}
    )
    assert yy.odc.geobox == xx.odc.geobox
    assert isinstance(xx.odc.geobox, GeoBox)

    yy = stac_load(
        [item],
        ["nir"],
        geobox=xx.odc.geobox,
        chunks={},
        stac_cfg={"*": {"warnings": "ignore"}},
    )
    assert yy.odc.geobox == xx.odc.geobox
    assert yy.odc.geobox == yy.nir.odc.geobox

    # Check automatic CRS/resolution
    yy = stac_load(
        [item],
        ["nir", "coastal"],
        chunks={},
        stac_cfg={"*": {"warnings": "ignore"}},
    )
    assert yy.odc.geobox.crs == "EPSG:32606"
    assert yy.odc.geobox.resolution.yx == (-10, 10)

    # test bbox overlaping with lon/lat
    with pytest.raises(ValueError):
        stac_load([item], ["nir"], bbox=(0, 0, 1, 1), lon=(0, 1), lat=(0, 1), chunks={})

    # test bbox overlaping with x/y
    with pytest.raises(ValueError):
        stac_load(
            [item],
            ["nir"],
            bbox=(0, 0, 1, 1),
            x=(0, 1000),
            y=(0, 1000),
            chunks={},
        )

    bbox = (0, 0, 1, 1)
    x1, y1, x2, y2 = bbox

    assert (
        stac_load(
            [item],
            ["nir"],
            crs="epsg:3857",
            resolution=10,
            chunks={},
            lon=(x1, x2),
            lat=(y1, y2),
        ).nir.odc.geobox
        == stac_load(
            [item],
            ["nir"],
            crs="epsg:3857",
            resolution=10,
            chunks={},
            bbox=bbox,
        ).nir.odc.geobox
    )

    geopolygon = shapely.geometry.box(*bbox)
    assert (
        stac_load(
            [item],
            ["nir"],
            crs="epsg:3857",
            resolution=10,
            chunks={},
            lon=(x1, x2),
            lat=(y1, y2),
        ).nir.odc.geobox
        == stac_load(
            [item],
            ["nir"],
            crs="epsg:3857",
            resolution=10,
            chunks={},
            geopolygon=geopolygon,
        ).nir.odc.geobox
    )


def test_group_items() -> None:
    def _mk(id: str, lon: float, datetime: str):
        gbox = GeoBox.from_bbox((lon - 0.1, 0, lon + 0.1, 1), shape=(100, 100))
        return mk_parsed_item([b_("b1", gbox)], datetime=datetime, id=id)

    # check no-op case first
    assert _group_items([], [], "time") == []
    assert _group_items([], [], "id") == []
    assert _group_items([], [], "solar_day") == []

    aa = _mk("a", 15 * 10, "2020-01-02T23:59Z")
    b1 = _mk("b1", 15 * 10 + 1, "2020-01-03T00:01Z")
    b2 = _mk("b2", 15 * 10 + 2, "2020-01-03T00:01Z")
    cc = _mk("c", 0, "2020-01-02T23:59Z")

    def _t(items, groupby, expect, lon=None, preserve_original_order=False) -> None:
        stac_items = [to_stac_item(item) for item in items]
        rr = ndeepmap(
            2,
            lambda idx: items[idx],
            _group_items(
                stac_items,
                items,
                groupby,
                lon=lon,
                preserve_original_order=preserve_original_order,
            ),
        )
        _expect = ndeepmap(2, lambda item: item.id, expect)
        _got = ndeepmap(2, lambda item: item.id, rr)

        assert _expect == _got

    # same order as input
    _t([aa, b1, b2], "id", [[aa], [b1], [b2]])
    _t([aa, b2, b1], "id", [[aa], [b2], [b1]])
    _t([b1, aa, b2], "id", [[b1], [aa], [b2]])
    _t([cc, aa, b1, b2], "id", [[cc], [aa], [b1], [b2]])

    _t([aa, b1, b2], "time", [[aa], [b1, b2]])
    _t([b1, aa, b2], "time", [[aa], [b1, b2]])

    # order within group is preserved
    _t([b2, aa, b1], "time", [[aa], [b2, b1]], preserve_original_order=True)
    _t([aa, cc, b1, b2], "time", [[aa, cc], [b1, b2]], preserve_original_order=True)

    _t([aa, b1, b2], "solar_day", [[aa, b1, b2]])
    _t([b1, aa, b2], "solar_day", [[aa, b1, b2]])
    _t([b2, aa, b1], "solar_day", [[aa, b1, b2]])
    _t([aa, b1, b2, cc], "solar_day", [[cc], [aa, b1, b2]])

    _t([aa, b1, b2, cc], "solar_day", [[aa, cc, b1, b2]], lon=150 + 1)

    # property based
    _t([aa, b1], "proj:epsg", [[aa, b1]])
    _t([b1, aa], "proj:epsg", [[aa, b1]])
    _t([aa, b1], "proj:transform", [[aa], [b1]])

    # custom callback
    _t(
        [aa, b1, b2, cc],
        lambda item, parsed, idx: idx % 2,
        [[aa, b2], [b1, cc]],
        preserve_original_order=True,
    )


def test_resolve_load_cfg() -> None:
    rlp = RasterLoadParams
    assert resolve_load_cfg({}) == {}

    item = mk_parsed_item(
        [
            b_("a", dtype="int8", nodata=-1),
            b_("b", dtype="float64", dims=("y", "x", "b")),
        ]
    )

    assert set(item.collection) == set([("a", 1), ("b", 1)])
    assert item.collection["a"].data_type == "int8"
    assert item.collection["b"].data_type == "float64"

    _bands = {n: b for (n, _), b in item.collection.bands.items()}
    assert isinstance(_bands["a"], RasterBandMetadata)
    assert isinstance(_bands["b"], RasterBandMetadata)

    cfg = resolve_load_cfg(_bands, resampling="average")
    assert cfg["a"] == rlp("int8", -1, resampling="average", meta=_bands["a"])
    assert cfg["b"] == rlp(
        "float64", None, resampling="average", dims=("y", "x", "b"), meta=_bands["b"]
    )

    cfg = resolve_load_cfg(
        _bands,
        resampling={"*": "mode", "b": "sum"},
        nodata=-999,
        dtype="int64",
    )
    assert cfg["a"] == rlp("int64", -999, resampling="mode", meta=_bands["a"])
    assert cfg["b"] == rlp(
        "int64", -999, resampling="sum", dims=("y", "x", "b"), meta=_bands["b"]
    )

    cfg = resolve_load_cfg(
        _bands,
        dtype={"a": "float32"},
    )
    assert cfg["a"] == rlp("float32", -1, meta=_bands["a"])
    assert cfg["b"] == rlp("float64", None, dims=_bands["b"].dims, meta=_bands["b"])
